﻿#include "pch.h"
#include "ClipToTextRenderer.h"
#include "Common/DirectXHelper.h"	// For ThrowIfFailed

using namespace ClipToText;

using namespace D2D1;
using namespace Microsoft::WRL;                 // for ComPtr
using namespace Platform;                       // for String
using namespace Windows::ApplicationModel;      // for Package

ClipToTextRenderer::ClipToTextRenderer(const std::shared_ptr<DeviceResources>& deviceResources) : 
    m_deviceResources(deviceResources)
{
    // Get font file
    ComPtr<IDWriteFactory> factory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Miramob.ttf";
    ComPtr<IDWriteFontFile> fontFile;
    DX::ThrowIfFailed(
        factory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Get font face
    ComPtr<IDWriteFontFace> fontFace;
    DX::ThrowIfFailed(
        factory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
                                1,
                                fontFile.GetAddressOf(),
                                0,
                                DWRITE_FONT_SIMULATIONS_NONE,
                                &fontFace)
        );

    // Create path geometry and open it
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreatePathGeometry(&m_clipGeometry)
        );

    ComPtr<ID2D1GeometrySink> geometrySink;
    DX::ThrowIfFailed(
        m_clipGeometry->Open(&geometrySink)
        );

    // Get glyph run outline ("CLIP")
    uint16 glyphIndices [] = { 0x0026, 0x002F, 0x002C, 0x0033 };
    float emSize = 96.0f;      // 72 points, arbitrary in this program

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(emSize,
                                     glyphIndices,
                                     nullptr,
                                     nullptr,
                                     ARRAYSIZE(glyphIndices),
                                     false,
                                     false,
                                     geometrySink.Get())
        );

    // Don't forget to close the geometry sink!
    geometrySink->Close();

    CreateDeviceDependentResources();
}

void ClipToTextRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::White), &m_whiteBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    m_needsRedraw = true;
}
void ClipToTextRenderer::ReleaseDeviceDependentResources()
{
    m_whiteBrush.Reset();
    m_stateBlock.Reset();
}

void ClipToTextRenderer::CreateWindowSizeDependentResources()
{
    m_needsRedraw = true;
}

void ClipToTextRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
bool ClipToTextRenderer::Render()
{
    if (!m_needsRedraw)
        return false;

    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size outputBounds = m_deviceResources->GetOutputBounds();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::DarkBlue));

    // Get the clip geometry bounds
    D2D_RECT_F geometryBounds;
    DX::ThrowIfFailed(
        m_clipGeometry->GetBounds(D2D1::IdentityMatrix(), &geometryBounds)
        );

    // Define transforms to center and scale clipping geometry
    Matrix3x2F orientationTransform = m_deviceResources->GetOrientationTransform2D();
    Matrix3x2F translateTransform = Matrix3x2F::Translation(SizeF(-geometryBounds.left, 
                                                                  -geometryBounds.top));

    float scaleHorz = outputBounds.Width / (geometryBounds.right - geometryBounds.left);
    float scaleVert = outputBounds.Height / (geometryBounds.bottom - geometryBounds.top);
    Matrix3x2F scaleTransform = Matrix3x2F::Scale(SizeF(scaleHorz, scaleVert));

    // Set the geometry for clipping
    ComPtr<ID2D1Layer> layer;
    DX::ThrowIfFailed(
        context->CreateLayer(&layer)
        );

    context->PushLayer(LayerParameters(InfiniteRect(), 
                                       m_clipGeometry.Get(), 
                                       D2D1_ANTIALIAS_MODE_PER_PRIMITIVE, 
                                       translateTransform * scaleTransform * orientationTransform), 
                       layer.Get());

    // Draw lines radiating from center
    translateTransform = Matrix3x2F::Translation(outputBounds.Width / 2, outputBounds.Height / 2);

    for (float angle = 0; angle < 360; angle += 1)
    {
        Matrix3x2F rotationTransform = Matrix3x2F::Rotation(angle);
        context->SetTransform(rotationTransform * translateTransform * orientationTransform);
        context->DrawLine(Point2F(0, 0), 
                          Point2F(outputBounds.Width / 2, outputBounds.Height / 2), 
                          m_whiteBrush.Get(), 2);
    }

    context->PopLayer();

    // We ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
    m_needsRedraw = false;
    return true;
}

