#include "pch.h"
#include "InterpolatableGeometrySink.h"

using namespace D2D1;
using namespace DirectX;

// IUnknown methods
ULONG STDMETHODCALLTYPE InterpolatableGeometrySink::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE InterpolatableGeometrySink::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE InterpolatableGeometrySink::QueryInterface(_In_ REFIID riid, _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(ID2D1SimplifiedGeometrySink))
    {
        *ppOutput = static_cast<ID2D1SimplifiedGeometrySink*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

// ID2D1SimplifiedGeometrySink methods
InterpolatableGeometrySink::InterpolatableGeometrySink(bool interpolate, float flatteningTolerance) : 
    m_refCount(0), 
    m_interpolate(interpolate),
    m_flatteningTolerance(flatteningTolerance)
{
}

void InterpolatableGeometrySink::SetFillMode(D2D1_FILL_MODE fillMode)
{
    m_pathGeometryData.FillMode = fillMode;
}

void InterpolatableGeometrySink::SetSegmentFlags(D2D1_PATH_SEGMENT vertexFlags)
{
    // Ignore
}

void InterpolatableGeometrySink::BeginFigure(D2D1_POINT_2F startPoint, 
                                             D2D1_FIGURE_BEGIN figureBegin)
{
    m_pathFigureData.StartPoint = startPoint;
    m_pathFigureData.FigureBegin = figureBegin;
    m_pathFigureData.Segments.clear();
    m_prevPoint = startPoint;
}

void InterpolatableGeometrySink::AddLines(const D2D1_POINT_2F *points, 
                                          UINT pointsCount)
{
    PathSegmentData polyLineSegment;
    polyLineSegment.IsBezier = false;

    if (!m_interpolate)
    {
        polyLineSegment.Points.assign(points, points + pointsCount);
    }
    else
    {
        for (UINT i = 0; i < pointsCount; i++)
        {
            D2D1_POINT_2F point = points[i];
            Interpolate(m_prevPoint, point, polyLineSegment.Points);
            m_prevPoint = point;
        }
    }

    m_pathFigureData.Segments.push_back(polyLineSegment);
}

void InterpolatableGeometrySink::AddBeziers(const D2D1_BEZIER_SEGMENT *beziers, 
                                            UINT beziersCount)
{
    PathSegmentData polyBezierSegment;
    polyBezierSegment.IsBezier = true;
    polyBezierSegment.Beziers.assign(beziers, beziers + beziersCount);
    m_pathFigureData.Segments.push_back(polyBezierSegment);
    m_prevPoint = beziers[beziersCount - 1].point3;
}

void InterpolatableGeometrySink::EndFigure(D2D1_FIGURE_END figureEnd)
{
    m_pathFigureData.FigureEnd = figureEnd;

    if (m_interpolate && figureEnd == D2D1_FIGURE_END_CLOSED)
    {
        PathSegmentData polyLineSegment;
        polyLineSegment.IsBezier = false;
        Interpolate(m_prevPoint, m_pathFigureData.StartPoint, polyLineSegment.Points);
        m_pathFigureData.Segments.push_back(polyLineSegment);
    }

    m_pathGeometryData.Figures.push_back(m_pathFigureData);
}

HRESULT InterpolatableGeometrySink::Close()
{
    // Assume that the class accessing the geometry sink knows what it's doing
    return S_OK;
}

// Methods for this implementation
PathGeometryData InterpolatableGeometrySink::GetPathGeometryData()
{
    return m_pathGeometryData;
}

void InterpolatableGeometrySink::Interpolate(D2D1_POINT_2F point0, 
                                             D2D1_POINT_2F point1, 
                                             std::vector<D2D1_POINT_2F>& points)
{
    XMVECTOR pt0 = XMVectorSet(point0.x, point0.y, 0, 0);
    XMVECTOR pt1 = XMVectorSet(point1.x, point1.y, 0, 0);
    XMVECTOR vector = pt1 - pt0;

    float length = XMVectorGetX(XMVector2Length(vector));

    if (length < m_flatteningTolerance)
    {
        points.push_back(point1);
    }
    else
    {
        int count = (int) (length / m_flatteningTolerance) + 1;
        XMVECTOR increment = vector / float(count);

        for (int i = 0; i < count; i++)
        {
            pt0 += increment;
            points.push_back(Point2F(XMVectorGetX(pt0), XMVectorGetY(pt0)));
        }
    }
}
