﻿#include "pch.h"
#include "OscillatingTextRenderer.h"
#include "Common/DirectXHelper.h"	// For ThrowIfFailed

using namespace OscillatingText;

using namespace D2D1;
using namespace DirectX;                        // for trig math
using namespace Microsoft::WRL;                 // for ComPtr
using namespace Platform;                       // for String
using namespace Windows::ApplicationModel;      // for Package

// Initializes D2D resources used for text rendering.
OscillatingTextRenderer::OscillatingTextRenderer(const std::shared_ptr<DeviceResources>& deviceResources) : 
	m_deviceResources(deviceResources)
{
    // Load font file
    ComPtr<IDWriteFactory> factory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Miramo.ttf";
    ComPtr<IDWriteFontFile> fontFile;
    DX::ThrowIfFailed(
        factory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;
    DX::ThrowIfFailed(
        factory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
        1,
        fontFile.GetAddressOf(),
        0,
        DWRITE_FONT_SIMULATIONS_NONE,
        &fontFace)
        );

    // Obtain glyph geometry in InterpolatableGeometrySink
    ComPtr<InterpolatableGeometrySink> interpolatableGeometrySink = new InterpolatableGeometrySink(true);
    uint16 glyphIndices [] = 
    { 
        // "HELLO WORLD"
        0x002B, 0x0028, 0x002F, 0x002F, 0x0032, 0x0003,
        0x003A, 0x0032, 0x0035, 0x002F, 0x0027
    };    
    float emSize = 144;

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(emSize,
                                     glyphIndices,
                                     nullptr,
        nullptr,
        ARRAYSIZE(glyphIndices),
        false,
        false,
        interpolatableGeometrySink.Get())
        );

    // Not really needed, but do it anyway
    interpolatableGeometrySink->Close();

    // Get two PathGeometryData objects
    m_srcPathGeometryData = interpolatableGeometrySink->GetPathGeometryData();
    m_dstPathGeometryData = interpolatableGeometrySink->GetPathGeometryData();

    // Generate a ID2D1PathGeometry objects from that geometry data
    m_textGeometry = m_srcPathGeometryData.GeneratePathGeometry(m_deviceResources->GetD2DFactory());

    // Get the geometry bounds
    DX::ThrowIfFailed(
        m_textGeometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

	CreateDeviceDependentResources();
}

void OscillatingTextRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::White), &m_whiteBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );
}
void OscillatingTextRenderer::ReleaseDeviceDependentResources()
{
    m_whiteBrush.Reset();
    m_stateBlock.Reset();
}

void OscillatingTextRenderer::CreateWindowSizeDependentResources()
{
    Windows::Foundation::Size outputBounds = m_deviceResources->GetOutputBounds();

    Matrix3x2F screenTranslation = Matrix3x2F::Translation(
        (outputBounds.Width - (m_geometryBounds.right + m_geometryBounds.left)) / 2,
        (outputBounds.Height - (m_geometryBounds.bottom + m_geometryBounds.top)) / 2
        );

    m_displayTransform = screenTranslation * m_deviceResources->GetOrientationTransform2D();
}

// Updates the text to be displayed.
void OscillatingTextRenderer::Update(DX::StepTimer const& timer)
{
    // 0 to 1 every 3 seconds
    float t =  float(fmod(timer.GetTotalSeconds(), 3) / 3);

    // Simusoidal oscillation between 1 and -1 every 3 seconds
    t = XMScalarSin(t * XM_2PI);
    float maxDeviation = (m_geometryBounds.bottom - m_geometryBounds.top) / 2 *(float) t;

    // lambda function to perform point translations
    auto Translate = [this, maxDeviation] (D2D1_POINT_2F point)
    {
        // vertical coordinate = 0 to 1 relative to top
        float vert = (point.y - m_geometryBounds.top) / (m_geometryBounds.bottom - m_geometryBounds.top);
        
        // vertical coordinate = 1 to 0 to 1 relative to top
        vert = 2 * (0.5f - vert) * (vert < 0.5f ? 1 : -1);

        // Equation of circle to make arcs
        vert = sqrt(1 - vert * vert);

        return Point2F(point.x + maxDeviation * vert, point.y);
    };

    for (size_t figIndex = 0; figIndex < m_srcPathGeometryData.Figures.size(); figIndex++)
    {
        PathFigureData& srcPathFigureData = m_srcPathGeometryData.Figures[figIndex];
        PathFigureData& dstPathFigureData = m_dstPathGeometryData.Figures[figIndex];

        dstPathFigureData.StartPoint = Translate(srcPathFigureData.StartPoint);

        for (size_t segIndex = 0; segIndex < srcPathFigureData.Segments.size(); segIndex++)
        {
            PathSegmentData& srcPathSegmentData = srcPathFigureData.Segments[segIndex];
            PathSegmentData& dstPathSegmentData = dstPathFigureData.Segments[segIndex];

            if (!srcPathSegmentData.IsBezier)
            {
                for (size_t ptIndex = 0; ptIndex < srcPathSegmentData.Points.size(); ptIndex++)
                {
                    dstPathSegmentData.Points[ptIndex] = Translate(srcPathSegmentData.Points[ptIndex]);
                }
            }
            else
            {
                for (size_t bezIndex = 0; bezIndex < srcPathSegmentData.Beziers.size(); bezIndex++)
                {
                    D2D1_BEZIER_SEGMENT& srcBezier = srcPathSegmentData.Beziers[bezIndex];
                    D2D1_BEZIER_SEGMENT& dstBezier = dstPathSegmentData.Beziers[bezIndex];

                    dstBezier.point1 = Translate(srcBezier.point1);
                    dstBezier.point2 = Translate(srcBezier.point2);
                    dstBezier.point3 = Translate(srcBezier.point3);
                }
            }
        }
    }

    m_textGeometry = m_dstPathGeometryData.GeneratePathGeometry(m_deviceResources->GetD2DFactory());
}

// Renders a frame to the screen.
void OscillatingTextRenderer::Render()
{
	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::DarkBlue));

    context->SetTransform(m_displayTransform);
    context->FillGeometry(m_textGeometry.Get(), m_whiteBrush.Get());

	// We ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());
}

