﻿#include "pch.h"
#include "RealTextEditorRenderer.h"
#include "Common/DirectXHelper.h"	// For ThrowIfFailed
#include "InterrogableGeometrySink.h"

using namespace RealTextEditor;

using namespace D2D1;
using namespace Microsoft::WRL;                 // for ComPtr
using namespace Platform;                       // for String
using namespace Windows::ApplicationModel;      // for Package

RealTextEditorRenderer::RealTextEditorRenderer(const std::shared_ptr<DeviceResources>& deviceResources) : 
	m_deviceResources(deviceResources),
    m_editMode(false)
{
    // Load font file
    ComPtr<IDWriteFactory> factory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Kooten.ttf";
    ComPtr<IDWriteFontFile> fontFile;

    DX::ThrowIfFailed(
        factory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;

    DX::ThrowIfFailed(
        factory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE, 
                                1, 
                                fontFile.GetAddressOf(), 
                                0, 
                                DWRITE_FONT_SIMULATIONS_NONE, 
                                &fontFace)
        );

    // Obtain glyph geometry in InterrogableGeometrySink
    ComPtr<InterrogableGeometrySink> interrogableGeometrySink = new InterrogableGeometrySink();
    uint16 glyphIndices[] = { 0x0027, 0x003B };     // "DX"
    float emSize = 768;                             // 8 inches

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(emSize, 
                                     glyphIndices, 
                                     nullptr, 
                                     nullptr, 
                                     ARRAYSIZE(glyphIndices), 
                                     false, 
                                     false, 
                                     interrogableGeometrySink.Get())
        );

    // Not really needed, but do it anyway
    interrogableGeometrySink->Close();

    // Get the PathGeometryData object
    m_pathGeometryData = interrogableGeometrySink->GetPathGeometryData();

    // Generate an ID2D1PathGeometry from that geometry data
    m_textGeometry = m_pathGeometryData.GeneratePathGeometry(m_deviceResources->GetD2DFactory());

    // Get the geometry bounds
    DX::ThrowIfFailed(
        m_textGeometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

	CreateDeviceDependentResources();
}

void RealTextEditorRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::DarkBlue), &m_darkBlueBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Red), &m_redBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Green), &m_greenBrush)
        );


    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    m_needsRedraw = true;
}
void RealTextEditorRenderer::ReleaseDeviceDependentResources()
{
    m_darkBlueBrush.Reset();


    m_redBrush.Reset();
    m_greenBrush.Reset();

    m_stateBlock.Reset();
}
void RealTextEditorRenderer::CreateWindowSizeDependentResources()
{
//    m_compositeTransform = 

    Windows::Foundation::Size outputBounds = m_deviceResources->GetOutputBounds();

    Matrix3x2F screenTranslation = Matrix3x2F::Translation(
        (outputBounds.Width - (m_geometryBounds.right + m_geometryBounds.left)) / 2,
        (outputBounds.Height - (m_geometryBounds.bottom + m_geometryBounds.top)) / 2
        );

    m_displayTransform = screenTranslation * m_deviceResources->GetOrientationTransform2D();

    // And get the inverse matrix for hit-testing
    m_inverseDisplayTransform = m_displayTransform;
    m_inverseDisplayTransform.Invert();

    m_needsRedraw = true;
}

// Updates the text to be displayed.
void RealTextEditorRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
bool RealTextEditorRenderer::Render()
{
    if (!m_needsRedraw)
        return false;

	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::LightGoldenrodYellow));
	context->SetTransform(m_displayTransform);

    if (!m_editMode)
    {
        context->FillGeometry(m_textGeometry.Get(), m_darkBlueBrush.Get());
    }
    else
    {
        context->DrawGeometry(m_textGeometry.Get(), m_blackBrush.Get(), 3.0f);

        for (const PathFigureData& pathFigureData : m_pathGeometryData.Figures)
        {
            float dotRadius = float(m_dotRadius);
            context->FillEllipse(Ellipse(pathFigureData.StartPoint, dotRadius, dotRadius), m_greenBrush.Get());
            D2D1_POINT_2F lastPoint = pathFigureData.StartPoint;

            for (const PathSegmentData& pathSegmentData : pathFigureData.Segments)
            {
                if (pathSegmentData.IsBezier)
                {
                    for (const D2D1_BEZIER_SEGMENT& bezier : pathSegmentData.Beziers)
                    {
                        context->FillEllipse(Ellipse(bezier.point1, dotRadius, dotRadius), m_redBrush.Get());
                        context->FillEllipse(Ellipse(bezier.point2, dotRadius, dotRadius), m_redBrush.Get());
                        context->FillEllipse(Ellipse(bezier.point3, dotRadius, dotRadius), m_greenBrush.Get());

                        // Draw lines from end points to control points
                        context->DrawLine(lastPoint, bezier.point1, m_redBrush.Get());
                        context->DrawLine(bezier.point2, bezier.point3, m_redBrush.Get());

                        // Save last point
                        lastPoint = bezier.point3;
                    }
                }
                else
                {
                    for (const D2D1_POINT_2F point : pathSegmentData.Points)
                    {
                        context->FillEllipse(Ellipse(point, dotRadius, dotRadius), m_greenBrush.Get());
                        lastPoint = point;
                    }
                }
            }
        }
    }

	// We ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());

    m_needsRedraw = false;
    return true;
}

void RealTextEditorRenderer::OnPointerPressed(unsigned int id, D2D1_POINT_2F pointerPos)
{
    // Get the iterator for pointerTracker
    auto const it = pointerTracker.find(id);

    // If this iterator already exists, remove it from the map
    if (it != pointerTracker.end())
    {
        OutputDebugString(L"OnPointerPressed without corresponding OnPointerReleased");
        pointerTracker.erase(it);
    }

    // If we're not in edit mode, switch to edit mode
    if (!m_editMode)
    {
        m_editMode = true;
        m_needsRedraw = true;
        return;
    }

    pointerPos = m_inverseDisplayTransform.TransformPoint(pointerPos);
    bool haveHit = false;
    PointerInfo pointerInfo;

    for (float offset = 1; offset <= m_dotRadius; offset++)
    {
        for (size_t figIndex = 0; figIndex < m_pathGeometryData.Figures.size(); figIndex++)
        {
            const PathFigureData& pathFigureData = m_pathGeometryData.Figures.at(figIndex);

            haveHit = HitTest(pointerPos, pathFigureData.StartPoint, offset);

            if (haveHit)
            {
                pointerInfo.FigureIndex = figIndex;
                pointerInfo.SegmentIndex = -1;
                break;
            }

            for (size_t segIndex = 0; segIndex < pathFigureData.Segments.size(); segIndex++)
            {
                const PathSegmentData& pathSegmentData = pathFigureData.Segments.at(segIndex);

                if (!pathSegmentData.IsBezier)
                {
                    for (size_t ptIndex = 0; ptIndex < pathSegmentData.Points.size(); ptIndex++)
                    {
                        D2D1_POINT_2F point = pathSegmentData.Points.at(ptIndex);
                        haveHit = HitTest(pointerPos, point, offset);

                        if (haveHit)
                        {
                            // Set two fields of the PointerInfo structure
                            pointerInfo.IsBezier = false;
                            pointerInfo.PointIndex = ptIndex;
                            break;
                        }
                    }
                }
                else
                {
                    for (size_t bezIndex = 0; bezIndex < pathSegmentData.Beziers.size(); bezIndex++)
                    {
                        const D2D1_BEZIER_SEGMENT& bezier = pathSegmentData.Beziers.at(bezIndex);
                        int hitPoint = 0;

                        if (HitTest(pointerPos, bezier.point1, offset))
                        {
                            hitPoint = 1;
                        }
                        else if (HitTest(pointerPos, bezier.point2, offset))
                        {
                            hitPoint = 2;
                        }
                        else if (HitTest(pointerPos, bezier.point3, offset))
                        {
                            hitPoint = 3;
                        }

                        haveHit = hitPoint > 0;

                        if (haveHit)
                        {
                            // Set three fields of the PointerInfo structure
                            pointerInfo.IsBezier = true;
                            pointerInfo.BezierIndex = bezIndex;
                            pointerInfo.PointIndex = hitPoint;
                            break;
                        }
                    }
                }

                if (haveHit)
                {
                    pointerInfo.FigureIndex = figIndex;
                    pointerInfo.SegmentIndex = segIndex;
                    break;
                }

            }   // segIndex loop

            if (haveHit)
                break;


        }   // figIndex loop

        if (haveHit)
        {
            // Set the remaining fields of the PointerInfo structure
            pointerInfo.LastPointerPosition = pointerPos;

            // Store in map with ID number
            pointerTracker.insert(std::make_pair(id, pointerInfo));
            break;
        }

    }   // offset loop

    // If not hits, switch out of edit mode
    if (!haveHit)
    {
        m_editMode = false;
        m_needsRedraw = true;
    }
}

void RealTextEditorRenderer::OnPointerMoved(unsigned int id, D2D1_POINT_2F pointerPos)
{
    // Get the iterator for pointerTracker
    auto const it = pointerTracker.find(id);

    // Never started a pointer operation, so skip
    if (it == pointerTracker.end())
        return;

    // Get the PointerInfo object for this finger
    pointerPos = m_inverseDisplayTransform.TransformPoint(pointerPos);
    PointerInfo& pointerInfo = it->second;
    TranslatePoint(pointerInfo, pointerPos);
    m_textGeometry = m_pathGeometryData.GeneratePathGeometry(m_deviceResources->GetD2DFactory());
    m_needsRedraw = true;
}

void RealTextEditorRenderer::OnPointerReleased(unsigned int id, D2D1_POINT_2F pointerPos)
{
    // Get the iterator for pointerTracker
    auto const it = pointerTracker.find(id);

    // Never started a pointer operation, so skip
    if (it == pointerTracker.end())
        return;

    // Get the PointerInfo object for this finger
    pointerPos = m_inverseDisplayTransform.TransformPoint(pointerPos);
    PointerInfo& pointerInfo = it->second;
    TranslatePoint(pointerInfo, pointerPos);
    m_textGeometry = m_pathGeometryData.GeneratePathGeometry(m_deviceResources->GetD2DFactory());

    // Remove from map
    pointerTracker.erase(it);
}

void RealTextEditorRenderer::TranslatePoint(PointerInfo& pointerInfo, D2D1_POINT_2F pointerPos)
{
    // Get amount to translate point, and save current position as last position
    float xDiff = pointerPos.x - pointerInfo.LastPointerPosition.x;
    float yDiff = pointerPos.y - pointerInfo.LastPointerPosition.y;
    pointerInfo.LastPointerPosition = pointerPos;

    // Define a lambda function to translate the point
    auto BumpUp = [xDiff, yDiff] (D2D1_POINT_2F& pt)
    {
        pt.x += xDiff;
        pt.y += yDiff;
    };

    PathFigureData& pathFigureData = m_pathGeometryData.Figures.at(pointerInfo.FigureIndex);

    // Check for figure start point
    if (pointerInfo.SegmentIndex == -1)
    {
        BumpUp(pathFigureData.StartPoint);
    }
    else
    {
        PathSegmentData& pathSegmentData = pathFigureData.Segments.at(pointerInfo.SegmentIndex);

        if (!pointerInfo.IsBezier)
        {
            BumpUp(pathSegmentData.Points.at(pointerInfo.PointIndex));
        }
        else
        {
            D2D1_BEZIER_SEGMENT& bezier = pathSegmentData.Beziers.at(pointerInfo.BezierIndex);

            switch (pointerInfo.PointIndex)
            {
            case 1: BumpUp(bezier.point1); break;
            case 2: BumpUp(bezier.point2); break;
            case 3: BumpUp(bezier.point3); break;
            }
        }
    }
}
    
