#include "pch.h"
#include "InteroperableGeometrySink.h"

using namespace DirectX;
using namespace Microsoft::WRL;
using namespace Windows::Foundation;
using namespace Windows::UI::Xaml::Media;

// IUnknown methods
ULONG STDMETHODCALLTYPE InteroperableGeometrySink::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE InteroperableGeometrySink::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE InteroperableGeometrySink::QueryInterface(_In_ REFIID riid, _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(ID2D1SimplifiedGeometrySink))
    {
        *ppOutput = static_cast<ID2D1SimplifiedGeometrySink*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

// ID2D1SimplifiedGeometrySink methods
InteroperableGeometrySink::InteroperableGeometrySink(bool interpolate, float flatteningTolerance) :
    m_refCount(0),
    m_interpolate(interpolate),
    m_flatteningTolerance(flatteningTolerance),
    pathGeometry(ref new PathGeometry())
{
}

void InteroperableGeometrySink::SetFillMode(D2D1_FILL_MODE fillMode)
{
    pathGeometry->FillRule = (FillRule) fillMode;
}

void InteroperableGeometrySink::SetSegmentFlags(D2D1_PATH_SEGMENT vertexFlags)
{
    // No Windows Runtime PathGeometry equivalent
}

void InteroperableGeometrySink::BeginFigure(D2D1_POINT_2F startPoint, D2D1_FIGURE_BEGIN figureBegin)
{
    pathFigure = ref new PathFigure();
    pathFigure->StartPoint = Point(startPoint.x, startPoint.y);
    pathFigure->IsFilled = figureBegin == D2D1_FIGURE_BEGIN_FILLED;
    lastPoint = startPoint;
}

void InteroperableGeometrySink::AddLines(const D2D1_POINT_2F *points, UINT pointsCount)
{
    PolyLineSegment^ polyLineSegment = ref new PolyLineSegment();

    for (UINT i = 0; i < pointsCount; i++)
    {
        D2D1_POINT_2F point = points[i];

        if (!m_interpolate)
        {
            polyLineSegment->Points->Append(Point(point.x, point.y));
        }
        else
        {
            Interpolate(lastPoint, point, polyLineSegment->Points);
        }

        lastPoint = point;
    }

    pathFigure->Segments->Append(polyLineSegment);
}

void InteroperableGeometrySink::AddBeziers(const D2D1_BEZIER_SEGMENT *beziers, UINT beziersCount)
{
    PolyBezierSegment^ polyBezierSegment = ref new PolyBezierSegment();

    for (UINT i = 0; i < beziersCount; i++)
    {
        D2D1_BEZIER_SEGMENT bezier = beziers[i];
        polyBezierSegment->Points->Append(Point(bezier.point1.x, bezier.point1.y));
        polyBezierSegment->Points->Append(Point(bezier.point2.x, bezier.point2.y));
        polyBezierSegment->Points->Append(Point(bezier.point3.x, bezier.point3.y));

        lastPoint = bezier.point3;
    }

    pathFigure->Segments->Append(polyBezierSegment);
}

void InteroperableGeometrySink::EndFigure(D2D1_FIGURE_END figureEnd)
{
    pathFigure->IsClosed = figureEnd == D2D1_FIGURE_END_CLOSED;

    if (m_interpolate && pathFigure->IsClosed)
    {
        PolyLineSegment^ polyLineSegment = ref new PolyLineSegment();
        Interpolate(lastPoint, D2D1::Point2F(pathFigure->StartPoint.X, pathFigure->StartPoint.Y), polyLineSegment->Points);
        pathFigure->Segments->Append(polyLineSegment);
    }

    pathGeometry->Figures->Append(pathFigure);
}

HRESULT InteroperableGeometrySink::Close()
{
    // Assume that the class accessing the geometry sink knows what it's doing
    return S_OK;
}

// Method for this implementation
PathGeometry^ InteroperableGeometrySink::GetPathGeometry()
{
    return pathGeometry;
}

void InteroperableGeometrySink::Interpolate(D2D1_POINT_2F point0, D2D1_POINT_2F point1, PointCollection^ points)
{
    XMVECTOR pt0 = XMVectorSet(point0.x, point0.y, 0, 0);
    XMVECTOR pt1 = XMVectorSet(point1.x, point1.y, 0, 0);
    XMVECTOR vector = pt1 - pt0;

    float length = XMVectorGetX(XMVector2Length(vector));

    if (length < m_flatteningTolerance)
    {
        points->Append(Point(point1.x, point1.y));
    }
    else
    {
        int count = (int) (length / m_flatteningTolerance) + 1;
        XMVECTOR increment = vector / float(count);

        for (int i = 0; i < count; i++)
        {
            pt0 += increment;
            points->Append(Point(XMVectorGetX(pt0), XMVectorGetY(pt0)));
        }
    }
}

